# -*- coding: utf-8 -*-
"""
Created on Tue Jun 05 17:57:50 2012

@author: Mike
"""


#import python modules
import numpy as np
import time

#import compiled fortran library
import finallib

import sys
import os
import atomwrite


#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
#np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [finallib.equilibcalcenergy(Pos + NormDir * x, M, L, Cut) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = finallib.equilibcalcenergy(Pos + NormDir * dx, M, L, Cut)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [finallib.equilibcalcenergy(Pos + NormDir * 2. * dx, M, L, Cut)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [finallib.equilibcalcenergy(Pos + NormDir * Dists[-1], M, L, Cut)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = finallib.equilibcalcenergy(Pos + NormDir * d3, M, L, Cut)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = finallib.equilibcalcenergy(Pos + NormDir * d3, M, L, Cut)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = finallib.equilibcalcenergyforces(Pos, M, L, Cut, Forces)   
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = finallib.equilibcalcenergyforces(Pos, M, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos

def InitSurfacePositions(L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Plates: (200,3) array of capacitor positions
"""
    #compute integer grid # of locations for cubic lattice
    NLat = 10
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat+1, dtype=float)
    #make the position array of the capacitor plates' atoms
    Surface=np.zeros((2*len(r)**2,3),float)
    #Empty vector to be filled with dummy names of the capacitor plates
    SurfNames=[]    
    j=0
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    #for x in 
    for x in r:
        for y in r:
            #The Plates positions are determined here such that the plates span the x and y dimensions of the entire box
            Surface[j]=np.array([x,y,0.],float) 
            j+=1
            SurfNames.append('Au')
            Surface[j]=np.array([x,y,L],float)
            j+=1
            SurfNames.append('Au')
                     
    return Surface,SurfNames

def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float)
    r = r + (L-r[NLat-1])/2. #Center r array
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = finallib.calcenergyforces(Pos, M, L, Cut, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))


def ZHist(Pos,L,bins,M):
    """Returns a two z-position histograms, one for the head 
       group molecules and one for the tail group molecules
Input:
    Pos: (N,3) array of atomic positions
    L: Box length
    m: number of bins
    M: atoms in molecule
Output:
    HGhist: Z-position histogram for the head group
    TGhist: Z-position histogram for the tail group
"""

    HGhist = np.zeros((len(bins)-1), int) #Initialize bin arrays
    TGhist = np.zeros((len(bins)-1), int)
   
    for i in range(len(Pos)): #Go over all atoms
        if i % M == 0: #Check to see if head group
            for j in range(len(bins)-1):
                #Check to see if hg in bin
                if Pos[i,2] >= bins[j] and Pos[i,2] < bins[j+1]:
                    #if so, add count to bin                    
                    HGhist[j]+=1
        
        else: #Else is a tail group, repeat process
            for j in range(len(bins)-1):
                if Pos[i,2] >= bins[j] and Pos[i,2] < bins[j+1]:
                    TGhist[j]+=1

    return HGhist,TGhist


####################################################################

# Global Variables
N=144;
L = 10.
dt=.001;
T=1.0;
rc=2.5;
RescaleFreq=200;
PdbFrequency = 50;
m = 20;
HistFrequency = 100;

#Initialize Surface Atoms
Surface,SurfNames = InitSurfacePositions(L)

#Adjust SAM molecule size
# M = 4, 6, 8 atoms per molecule
#Tail group length = M - 1
M=4
    
#Construct SAM name array
NamesA = []
namecount=0
while namecount < N:
    NamesA.append('S') #First add head group "Thiol"
    namecount = namecount + 1
    tailcount = 1
    while tailcount <= M - 1:
        NamesA.append('C') #Then add tail group carbons
        namecount = namecount + 1
        tailcount = tailcount + 1
Names = NamesA+SurfNames #Combine Names with surface atom names                
Pdb = atomwrite.pdbfile("anim.pdb", L, AtomNames = Names) 
    
#Initialize atom position in lattice and minimize energy
Pos = InitPositions(N,L)
PE, Pos = ConjugateGradient(Pos, dx=.001, EFracTolLS=1.e-8, EFracTolCG=1.e-10)
Vel = InitVelocities(N,T)
Accel = InitAccel(Pos)
        
#Equilibration Step
t_equil = 0
while t_equil*dt <= 10.:
    Pos, Vel, Accel, KEnergy, PEnergy = finallib.equilibvvintegrate(Pos, Vel, Accel, M, L, rc, dt)
    
    #Check Total Energy        
    Tot = PEnergy + KEnergy
    #print "%.f       %.3f     %.3f     %.3f" %(t_equil,KEnergy,PEnergy,Tot)                        
    
    #Write atoms to PDB file
    if t_equil % PdbFrequency ==0:
        PosWrite = np.vstack((Pos,Surface)) #Combine Position and Surf coords
        Pdb.write(PosWrite)
                
    #Rescale velocity every 200 steps
    if t_equil % RescaleFreq == 0:
        Vel = RescaleVelocities(Vel,T)
                
    t_equil = t_equil + 1

#Initialize Histogram operators
dz=L/float(m) #bin size dz
bins=[(i*dz) for i in range(m+1)] #Bin spread
HGhist_counter = np.zeros((len(bins)-1), int) #Initialize Histogram Counters
TGhist_counter = np.zeros((len(bins)-1), int)
HGhist_5_counter = np.zeros((len(bins)-1), int) #Initialize Histogram Counters
TGhist_5_counter = np.zeros((len(bins)-1), int)
HGhist_10_counter = np.zeros((len(bins)-1), int) #Initialize Histogram Counters
TGhist_10_counter = np.zeros((len(bins)-1), int)
HGhist_20_counter = np.zeros((len(bins)-1), int) #Initialize Histogram Counters
TGhist_20_counter = np.zeros((len(bins)-1), int)
H_counter = 0;            
H_5_counter = 0;
H_10_counter = 0;
H_20_counter = 0;

#Initialize Energy Array
EArray = np.zeros(20001)

#Adsorption Step
t_ads = 0
while t_ads*dt <= 70.:
    
    Pos, Vel, Accel, KEnergy, PEnergy = finallib.vvintegrate(Pos, Vel, Accel, M, L, rc, dt)
    Tot = PEnergy + KEnergy
    #print "%.f       %.3f     %.3f     %.3f" %(t_ads,KEnergy,PEnergy,Tot)                        
    
    #Compute histogram after time t = 5
    if 5. <= t_ads*dt < 6.:
        if t_ads % HistFrequency == 0.:
            HGhist5,TGhist5 = ZHist(Pos,L,bins,M)
            HGhist_5_counter = HGhist_5_counter + HGhist5
            TGhist_5_counter = TGhist_5_counter + TGhist5
            H_5_counter = H_5_counter + 1
            
        
    #Compute histogram after time t = 10
    if 10. <= t_ads*dt < 11.:
        if t_ads % HistFrequency == 0.:
            HGhist10,TGhist10 = ZHist(Pos,L,bins,M)
            HGhist_10_counter = HGhist_10_counter + HGhist10
            TGhist_10_counter = TGhist_10_counter + TGhist10
            H_10_counter = H_10_counter + 1   

    #Compute histogram after time t = 20
    if 20. <= t_ads*dt < 21.:
        if t_ads % HistFrequency == 0.:
            HGhist20,TGhist20 = ZHist(Pos,L,bins,M)
            HGhist_20_counter = HGhist_20_counter + HGhist20
            TGhist_20_counter = TGhist_20_counter + TGhist20
            H_20_counter = H_20_counter + 1   
    
    #Write Positions to Pdb file for Chimera Visualization
    if t_ads % PdbFrequency == 0:
        #Combine SAM positions with surface atom positions                
        PosWrite = np.vstack((Pos,Surface))
        Pdb.write(PosWrite)

    #Store total energy values   
    if t_ads > 50000:
        EArray[t_ads - 50000] = Tot
        
    #Rescale velocity every 200 steps
    if t_ads % RescaleFreq == 0:
        Vel = RescaleVelocities(Vel,T)
        
    t_ads = t_ads + 1

#Calculate mean total energy for last 20 time units
EArrayMean = EArray.mean()
#Calculate the target kinetic energy as difference between average
#total energy and the current potential energy value
TargetKE = EArrayMean - PEnergy 
#recenter to zero net momentum (assuming all masses same)   
Vel = Vel - Vel.mean(axis=0)
#find velocity scale factor from ratios of velocity terms
#V^2 for the target kinetic energy
Vtargetsquared=2.0/N*(TargetKE)
#V^2 for the current kinetic energy
Vnowsquared=np.sum(np.mean(Vel*Vel,axis=0))
#Calculate velocity scale factor
VScale=np.sqrt(Vtargetsquared/Vnowsquared)
#compute new velocity
Vel = Vel * VScale
       
#Production Run
t_prod = 0
while t_prod * dt <= 100.:
    Pos, Vel, Accel, KEnergy, PEnergy = finallib.vvintegrate(Pos, Vel, Accel, M, L, rc, dt)
    
    if t_prod % HistFrequency == 0:
        HGhist,TGhist = ZHist(Pos,L,bins,M)
        HGhist_counter = HGhist_counter + HGhist
        TGhist_counter = TGhist_counter + TGhist
        H_counter = H_counter + 1
    
    t_prod = t_prod + 1

#Compute density histogram for t = 5 to 6
HGhist5density = HGhist_5_counter/(H_5_counter*L*L*L*dz)
TGhist5density = TGhist_5_counter/(H_5_counter*L*L*L*dz)
#Compute density histogram for t = 10 to 11
HGhist10density = HGhist_10_counter/(H_5_counter*L*L*L*dz)
TGhist10density = TGhist_10_counter/(H_5_counter*L*L*L*dz)
#Compute density histogram for t = 20 to 21
HGhist20density = HGhist_20_counter/(H_5_counter*L*L*L*dz)
TGhist20density = TGhist_20_counter/(H_5_counter*L*L*L*dz)
#Compute density histogram for production run
HGProdDensity = HGhist_counter/(H_counter*L*L*L*dz)
TGProdDensity = TGhist_counter/(H_counter*L*L*L*dz)

print "T = 5  Histogram"
print "(Min,Max)        HG Density      TG Density"
for i in range(len(bins)-1):
    print "(%.1f, %.1f)       %.5f         %.5f" %(bins[i],bins[i+1],HGhist5density[i],TGhist5density[i])      

print "T = 10  Histogram"
print "(Min,Max)        HG Density      TG Density"
for i in range(len(bins)-1):
    print "(%.1f, %.1f)       %.5f         %.5f" %(bins[i],bins[i+1],HGhist10density[i],TGhist10density[i])      

print "T = 20  Histogram"
print "(Min,Max)        HG Density      TG Density"
for i in range(len(bins)-1):
    print "(%.1f, %.1f)       %.5f         %.5f" %(bins[i],bins[i+1],HGhist20density[i],TGhist20density[i])      

print "Production  Histogram"
print "(Min,Max)        HG Density      TG Density"
for i in range(len(bins)-1):
    print "(%.1f, %.1f)       %.5f         %.5f" %(bins[i],bins[i+1],HGProdDensity[i],TGProdDensity[i])      
    
